@@ -3,8 +3,6 @@
  *
  * @APPLE_LICENSE_HEADER_START@
  * 
- * Copyright (c) 1999-2003 Apple Computer, Inc.  All Rights Reserved.
- * 
  * This file contains Original Code and/or Modifications of Original Code
  * as defined in and that are subject to the Apple Public Source License
  * Version 2.0 (the 'License'). You may not use this file except in
@@ -28,8 +26,9 @@
 //
 
 // SCSI Architecture Model Family includes
-#include <IOKit/scsi-commands/SCSICommandDefinitions.h>
-#include <IOKit/scsi-commands/IOSCSIReducedBlockCommandsDevice.h>
+#include <IOKit/scsi/SCSICommandDefinitions.h>
+
+#include "IOSCSIReducedBlockCommandsDevice.h"
 #include "SCSIReducedBlockCommands.h"
 
 
